module extexample03
use forpy_mod
use iso_c_binding
implicit none

! You need to declare exactly one PythonModule and PythonMethodTable
! at Fortran module level
type(PythonModule), save :: mod_def
type(PythonMethodTable), save :: method_table

CONTAINS

! Initialisation function for Python 3
! called when importing module
! must use bind(c, name="PyInit_<module name>")
! return value must be type(c_ptr), use the return value of PythonModule%init
function PyInit_extexample03() bind(c, name="PyInit_extexample03") result(m)
  type(c_ptr) :: m
  m = init()
end function

function init() result(m)
  type(c_ptr) :: m
  integer :: ierror
  type(object) :: pi

  ierror = forpy_initialize()

  call method_table%init(3) ! module shall have 1 method

  ! must add function print_args to method table to be able to use it in Python

  call method_table%add_method("print_args", &                  ! method name
                               "Prints arguments and keyword arguments", &  !doc-string
                               METH_VARARGS + METH_KEYWORDS, &  ! this method takes arguments AND keyword arguments
                               c_funloc(print_args))            ! address of Fortran function to add

  call method_table%add_method("getrij", &                  ! method name
                               "Gets interparticle distance", &  !doc-string
                               METH_VARARGS + METH_KEYWORDS, &  ! this method takes arguments AND keyword arguments
                               c_funloc(getrij))            ! address of Fortran function to add


 call method_table%add_method("function_add", &                  ! method name
                               "Adds numbers", &  !doc-string
                               METH_VARARGS + METH_KEYWORDS, &  ! this method takes arguments AND keyword arguments
                               c_funloc(function_add))            ! address of Fortran function to add

  m = mod_def%init("extexample03", "A Python extension with a method and a member.", method_table)

  ! Example: Numerical constant as member of module
  ierror = cast(pi, 3.141592653589793d0)
  ierror = mod_def%add_object("pi", pi)
  call pi%destroy
end function

! Implementation of our Python method
!
! Corresponding Python method shall allow arguments and keyword arguments
! -> We need 3 "type(c_ptr), value" arguments
! First arg is c_ptr to module, second is c_ptr to argument tuple
! third is c_ptr to keyword argument dict
! Return value must be type(c_ptr)
! bind(c) attribute to make sure that C calling conventions are used
function print_args(self_ptr, args_ptr, kwargs_ptr) result(r) bind(c)
  type(c_ptr), value :: self_ptr
  type(c_ptr), value :: args_ptr
  type(c_ptr), value :: kwargs_ptr
  type(c_ptr) :: r

  type(tuple) :: args
  type(dict) :: kwargs
  type(NoneType) :: retval
  integer :: ierror

  ! use unsafe_cast_from_c_ptr to cast from c_ptr to tuple/dict
  call unsafe_cast_from_c_ptr(args, args_ptr)
  call unsafe_cast_from_c_ptr(kwargs, kwargs_ptr)

  if (is_null(kwargs)) then
    ! This is a check if keyword argument were passed to this function.
    ! If is_null(kwargs), kwargs is not a valid Python object, therefore
    ! we initialise it as an empty dict
    ierror = dict_create(kwargs)
  endif

  ierror = print_py(args)
  ierror = print_py(kwargs)

  ! You always need to return something, at least None
  ierror = NoneType_create(retval)
  r = retval%get_c_ptr() ! need return value as c_ptr

  call args%destroy
  call kwargs%destroy

end function


function getrij(self_ptr, args_ptr, kwargs_ptr) result(r) bind(c)
  type(c_ptr), value :: self_ptr
  type(c_ptr), value :: args_ptr
  type(c_ptr), value :: kwargs_ptr
  type(c_ptr) :: r

  type(tuple) :: args
  type(dict) :: kwargs
  type(NoneType) :: retval
  integer :: ierror

  ! use unsafe_cast_from_c_ptr to cast from c_ptr to tuple/dict
  call unsafe_cast_from_c_ptr(args, args_ptr)
  call unsafe_cast_from_c_ptr(kwargs, kwargs_ptr)

  if (is_null(kwargs)) then
    ! This is a check if keyword argument were passed to this function.
    ! If is_null(kwargs), kwargs is not a valid Python object, therefore
    ! we initialise it as an empty dict
    ierror = dict_create(kwargs)
  endif

  ierror = print_py(args)
  ierror = print_py(kwargs)

  ! You always need to return something, at least None
  ierror = NoneType_create(retval)
  r = retval%get_c_ptr() ! need return value as c_ptr

  call args%destroy
  call kwargs%destroy

end function

function function_add(self_ptr, args_ptr) result(r) bind(c)
  type(c_ptr), value :: self_ptr
  type(c_ptr), value :: args_ptr
  !type(c_ptr), value :: kwargs_ptr
  type(c_ptr) :: r

  type(tuple) :: args
  !type(dict) :: kwargs
  type(NoneType) :: retval
  integer :: ierror

  ! use unsafe_cast_from_c_ptr to cast from c_ptr to tuple/dict
  call unsafe_cast_from_c_ptr(args, args_ptr)
  !call unsafe_cast_from_c_ptr(kwargs, kwargs_ptr)

  !if (is_null(kwargs)) then
    ! This is a check if keyword argument were passed to this function.
    ! If is_null(kwargs), kwargs is not a valid Python object, therefore
    ! we initialise it as an empty dict
  !  ierror = dict_create(kwargs)
  !endif

  ierror = print_py(args)
  !ierror = print_py(kwargs)
  ! You always need to return something, at least None
  ierror = NoneType_create(retval)
  r = retval%get_c_ptr() ! need return value as c_ptr

  call args%destroy
  !call kwargs%destroy

end function


end module
